import unittest

from ligo.gracedb import rest
import pytest
from pytest_socket import disable_socket

from .. import app


@pytest.fixture(scope='session', autouse=True)
def celeryconf():
    new_conf = dict(
        broker_url='memory://',
        result_backend='cache+memory://',
        gcn_bind_address='127.0.0.1',
        gcn_bind_port=53410,
        gcn_remote_address='127.0.0.1',
        task_always_eager=True,
        task_eager_propagates=True,
        lvalert_host='lvalert.invalid',
        gracedb_host='gracedb.invalid'
    )
    tmp = {key: app.conf[key] for key in new_conf.keys()}
    app.conf.update(new_conf)
    yield
    app.conf.update(tmp)


@pytest.fixture(autouse=True)
def fake_gracedb_client(monkeypatch):
    mock_client = unittest.mock.create_autospec(rest.GraceDb)
    mock_client.service_url = 'https://gracedb.invalid/api/'
    monkeypatch.setattr('gwcelery.tasks.gracedb.client', mock_client)
    yield


def pytest_runtest_setup():
    disable_socket()



from unittest.mock import Mock, patch

from ligo.gracedb import rest
import pkg_resources
import pytest

from ..tasks import orchestrator
from . import resource_json


# @patch('gwcelery.tasks.circulars.create_circular')
# @patch('gwcelery.tasks.skymaps.annotate_fits')
# @patch('gwcelery.tasks.bayestar.bayestar')
# @patch('gwcelery.tasks.gcn.send.delay')
# def test_handle_voevent(mock_send, mock_bayestar, mock_annotate_fits,
#                         mock_create_circular):
#     """Test dispatch of a VOEvent message."""
#     # Test LVAlert payload.
#     alert = resource_json(__name__, 'data/lvalert_voevent.json')
#
#     # text = alert['object']['text']
#
#     # Run function under test
#     orchestrator.handle(alert)
#
#     # Check that the correct tasks were dispatched.
#     mock_annotate_fits.assert_not_called()
#     mock_bayestar.assert_not_called()
#     mock_create_circular.assert_not_called()
#     # FIXME: temporarily disable sending GCNs as per P. Brady request
#     mock_send.assert_not_called()  # mock_send.assert_called_once_with(text)


@pytest.mark.parametrize(
    'group,other_group,annotate,other_annotate', [
        [
            'CBC', 'Burst',
            'gwcelery.tasks.orchestrator.annotate_cbc_superevent.run',
            'gwcelery.tasks.orchestrator.annotate_burst_superevent.run'
        ],
        [
            'Burst', 'CBC',
            'gwcelery.tasks.orchestrator.annotate_burst_superevent.run',
            'gwcelery.tasks.orchestrator.annotate_cbc_superevent.run'
        ]
    ])
def test_handle_superevent(monkeypatch, group, other_group,
                           annotate, other_annotate):
    """Test a superevent is dispatched to the correct annotation task based on
    its preferred event's search group."""
    alert = {
        'alert_type': 'new',
        'object': {
            'superevent_id': 'S1234'
        }
    }

    def get_superevent(superevent_id):
        assert superevent_id == 'S1234'
        return {'preferred_event': 'G1234'}

    def get_event(graceid):
        assert graceid == 'G1234'
        return {'group': group}

    mock_annotate = Mock()
    mock_other_annotate = Mock()

    monkeypatch.setattr('gwcelery.tasks.gracedb.get_superevent.run',
                        get_superevent)
    monkeypatch.setattr('gwcelery.tasks.gracedb.get_event.run',
                        get_event)
    monkeypatch.setattr(annotate, mock_annotate)
    monkeypatch.setattr(other_annotate, mock_other_annotate)

    # Run function under test
    orchestrator.handle_superevent(alert)

    mock_annotate.assert_called_once_with('G1234', 'S1234')

    # FIXME: The assertion below will fail in the unit tests because of an
    # issue in Celery with Ignore semipredicates in eager mode. However, it
    # is fine when running live.
    # See https://github.com/celery/celery/issues/4868.

    # mock_other_annotate.assert_not_called()


def mock_download(filename, graceid, *args, **kwargs):
    assert graceid == 'T250822'
    if filename == 'coinc.xml':
        return pkg_resources.resource_string(__name__, 'data/coinc.xml')
    elif filename == 'psd.xml.gz':
        return pkg_resources.resource_string(__name__, 'data/psd.xml.gz')
    else:
        raise ValueError


@patch('gwcelery.tasks.gracedb.download.run', mock_download)
@patch('gwcelery.tasks.em_bright.classifier.run')
@patch('gwcelery.tasks.bayestar.localize.run')
@patch('gwcelery.tasks.gracedb.client', autospec=rest.GraceDb)
def test_handle_cbc_event(mock_gracedb, mock_localize, mock_classifier):
    """Test that an LVAlert message for a newly uploaded PSD file triggers
    BAYESTAR."""
    alert = resource_json(__name__, 'data/lvalert_psd.json')
    orchestrator.handle_cbc_event(alert)
    mock_localize.assert_called_once()
    mock_classifier.assert_called_once()


@patch('gwcelery.tasks.gracedb.download.run', mock_download)
@patch('gwcelery.tasks.em_bright.classifier.run')
@patch('gwcelery.tasks.bayestar.localize.run')
@patch('gwcelery.tasks.gracedb.client', autospec=rest.GraceDb)
def test_handle_cbc_event_ignored(mock_gracedb, mock_localize,
                                  mock_classifier):
    """Test that unrelated LVAlert messages do not trigger BAYESTAR."""
    alert = resource_json(__name__, 'data/lvalert_detchar.json')
    orchestrator.handle_cbc_event(alert)
    mock_localize.assert_not_called()
    mock_classifier.assert_not_called()


@patch('gwcelery.tasks.raven.coincidence_search')
def test_handle_superevent_creation(mock_raven_coincidence_search):
    """Test dispatch of an LVAlert message for a superevent creation."""
    # Test LVAlert payload.
    alert = resource_json(__name__, 'data/lvalert_superevent_creation.json')

    # Run function under test
    orchestrator.handle_superevents_externaltriggers(alert)

    # Check that the correct tasks were dispatched.
    mock_raven_coincidence_search.assert_called_once_with('S180616h',
                                                          alert['object'])



"""Routing of LVAlert messages to other tasks."""
import json
from urllib.error import URLError

from celery import group
from celery.exceptions import Ignore
from ligo.gracedb.rest import HTTPError

from ..celery import app
from . import bayestar
from . import circulars
from .core import identity
from . import em_bright
from . import gracedb
from . import lvalert
from . import raven
from . import skymaps


@lvalert.handler('superevent',
                 'test_superevent',
                 shared=False)
def handle_superevent(alert):
    """Schedule annotations for new superevents.

    This task calls
    :func:`~gwcelery.tasks.orchestrator.annotate_cbc_superevent` after a
    timeout specified by the :obj:`~gwcelery.celery.Base.orchestrator_timeout`
    configuration variable.
    """

    if alert['alert_type'] != 'new':
        return

    superevent_id = alert['object']['superevent_id']
    (
        get_preferred_event.s(superevent_id)
        |
        group(
            continue_if_group_is.s('CBC')
            |
            annotate_cbc_superevent.s(superevent_id),

            continue_if_group_is.s('Burst')
            |
            annotate_burst_superevent.s(superevent_id)
        )
    ).apply_async(countdown=app.conf['orchestrator_timeout'])


@lvalert.handler('cbc_gstlal',
                 'cbc_pycbc',
                 'cbc_mbtaonline',
                 'test_gstlal',
                 'test_pycbc',
                 'test_mbtaonline',
                 shared=False)
def handle_cbc_event(alert):
    """Peform annotations for CBC events that depend on pipeline-specific
    matched-filter parameter estimates, including preliminary sky localization
    with BAYESTAR (via the :meth:`gwcelery.tasks.bayestar.localize` task) and
    preliminary source classification (via the
    :meth:`gwcelery.tasks.em_bright.em_bright` task)."""

    # Only handle alerts for the upload of a PSD file.
    if alert['alert_type'] != 'update' or alert.get('file') != 'psd.xml.gz':
        return

    graceid = alert['uid']

    (
        group(
            gracedb.download.s('coinc.xml', graceid),
            gracedb.download.s('psd.xml.gz', graceid)
        )
        |
        # FIXME: group(A, B) | group(C, D) does not pass the results from
        # tasks A and B to tasks C and D without this.
        identity.s()
        |
        group(
            bayestar.localize.s(graceid)
            |
            gracedb.upload.s(
                'bayestar.fits.gz', graceid,
                'sky localization complete', ['sky_loc', 'lvem']
            )
            |
            gracedb.create_label.si('SKYMAP_READY', graceid),

            em_bright.classifier.s(graceid)
            |
            gracedb.upload.s(
                'source_classification.json', graceid,
                'source classification complete', ['em_bright', 'lvem']
            )
            |
            gracedb.create_label.si('EMBRIGHT_READY', graceid)
        )
    ).delay()


@lvalert.handler('superevent',
                 'external_fermi',
                 'external_fermi_grb',
                 'external_grb',
                 'external_snews',
                 'external_snews_supernova',
                 'external_swift',
                 shared=False)
def handle_superevents_externaltriggers(alert):
    """Parse an LVAlert message related to superevents/external triggers and
    dispatch it to other tasks."""
    # Determine GraceDb ID
    graceid = alert['uid']

    if alert['alert_type'] == 'new':
        raven.coincidence_search(graceid, alert['object']).delay()


@app.task(autoretry_for=(HTTPError, URLError, TimeoutError),
          default_retry_delay=20.0, retry_backoff=True,
          retry_kwargs=dict(max_retries=500), shared=False)
def download(*args, **kwargs):
    """Download a file from GraceDb.

    This works just like :func:`gwcelery.tasks.gracedb.download`, except that
    it is retried for both `TimeoutError` and `URLError`. In particular, it
    will be retried for 404 (not found) errors.

    FIXME: Is there a way to change retry settings when calling a task,
    instead of creating an entirely new task?"""
    return gracedb.download(*args, **kwargs)


@app.task(autoretry_for=(URLError, TimeoutError), default_retry_delay=1.0,
          retry_backoff=True, retry_backoff_max=10.0,
          retry_kwargs=dict(max_retries=10), shared=False)
def get_preferred_event(superevent_id):
    """Determine preferred event for a superevent by querying GraceDb.

    This works just like :func:`gwcelery.tasks.gracedb.get_superevent`, except
    that it returns only the preferred event, and not the entire GraceDb JSON
    response."""
    return gracedb.get_superevent(superevent_id)['preferred_event']


@gracedb.task
def continue_if_group_is(graceid, group):
    """Continue processing if an event's group matches `group`, else halt
    the rest of the canvas."""
    if gracedb.get_event(graceid)['group'].lower() == group.lower():
        return graceid
    else:
        raise Ignore('This is not a {} event.'.format(group))


@gracedb.task
def create_voevent_for_em_bright(em_bright_json, *args, **kwargs):
    """Create a VOEvent record from an EM bright JSON file."""
    data = json.loads(em_bright_json)
    return gracedb.create_voevent(*args, **kwargs,
                                  ProbHasNS=0.01 * data['Prob NS2'],
                                  ProbHasRemnant=0.01 * data['Prob EMbright'])


@app.task(ignore_result=True, shared=False)
def annotate_burst_superevent(preferred_event_id, superevent_id):
    """Perform annotations for a superevent whose preferred event is a
    Burst."""
    (
        gracedb.create_voevent.s(superevent_id, 'preliminary')
        |
        group(
            circulars.create_circular.si(superevent_id),
            gracedb.create_label.si('GCN_PRELIM_SENT', superevent_id)
        )
    ).delay()


@app.task(ignore_result=True, shared=False)
def annotate_cbc_superevent(preferred_event_id, superevent_id):
    """Perform annotations for a superevent whose preferred event is a CBC."""
    (
        download.s('bayestar.fits.gz', preferred_event_id)
        |
        group(
            gracedb.upload.s(
                'bayestar.fits.gz',
                superevent_id,
                message='BAYESTAR localization copied from {}'.format(
                    preferred_event_id),
                tags=['sky_loc', 'lvem']
            )
            |
            gracedb.create_label.si('SKYMAP_READY', superevent_id),

            skymaps.annotate_fits(
                'bayestar.fits.gz',
                'bayestar',
                superevent_id,
                ['sky_loc', 'lvem']
            )
        )
        |
        download.si('source_classification.json', preferred_event_id)
        |
        group(
            create_voevent_for_em_bright.s(
                superevent_id, 'preliminary',
                skymap_type='BAYESTAR',
                skymap_filename='bayestar.fits.gz',
                skymap_image_filename='bayestar.png'
            )
            |
            gracedb.create_label.si('GCN_PRELIM_SENT', superevent_id),

            gracedb.upload.s(
                'source_classification.json',
                superevent_id,
                message='Source classification copied from {}'.format(
                    preferred_event_id),
                tags=['em_bright', 'lvem']
            )
            |
            gracedb.create_label.si('EMBRIGHT_READY', superevent_id),
        )
        |
        circulars.create_circular.si(superevent_id)
        |
        gracedb.upload.s(
            'circular.txt',
            superevent_id,
            'Automated circular'
        )
    ).delay()
